from indra.statements import *
from genewalk.nx_mg_assembler import IndraNxMgAssembler


def test_indra_assembly():
    braf = Agent('BRAF', db_refs={'HGNC': '1097', 'UP': 'P15056'})
    kras = Agent('KRAS', db_refs={'HGNC': '6407', 'UP': 'P01116'})
    genes = [{'HGNC': agent.db_refs['HGNC'],
              'HGNC_SYMBOL': agent.name,
              'UP': agent.db_refs['UP']} for agent in [braf, kras]]
    stmts = [Phosphorylation(kras, braf)]
    mg = IndraNxMgAssembler(genes, stmts)
    assert mg.graph
    assert mg.graph.nodes['KRAS']
    assert len(mg.graph.edges('BRAF')) > 5


import os
import gzip
import shutil
import logging
import urllib.request
import pandas
from indra.databases import hgnc_client

logger = logging.getLogger('genewalk.resources')


class ResourceManager(object):
    """Class to manage the download, caching and access of resource files."""
    def __init__(self, base_folder=None):
        self.base_folder = base_folder if base_folder else \
            os.path.join(os.path.expanduser('~'), 'genewalk')
        self.resource_folder = self._get_resource_folder()
        logger.info('Using %s as resource folder.' % self.resource_folder)

    def get_go_obo(self):
        fname = os.path.join(self.resource_folder, 'go.obo')
        if not os.path.exists(fname):
            url = 'http://snapshot.geneontology.org/ontology/go.obo'
            download_url(url, fname)
        return fname

    def get_goa_gaf(self):
        fname = os.path.join(self.resource_folder, 'goa_human.gaf')
        if not os.path.exists(fname):
            url_goa = ('http://geneontology.org/gene-associations/'
                       'goa_human.gaf.gz')
            download_gz(fname, url_goa)
        return fname

    def get_pc(self):
        fname_current = os.path.join(self.resource_folder,
                                     'PathwayCommons12.All.hgnc_current.sif')
        if not os.path.exists(fname_current):
            fname = os.path.join(self.resource_folder,
                                 'PathwayCommons12.All.hgnc.sif')
            if not os.path.exists(fname):
                url_pc = ('http://www.pathwaycommons.org/archives/PC2/v12/'
                          'PathwayCommons12.All.hgnc.sif.gz')
                download_gz(fname, url_pc)
            self._replace_outdated_hgnc_symbols(fname,fname_current)
        return fname_current

    def get_mgi_entrez(self):
        fname = os.path.join(self.resource_folder, 'MGI_EntrezGene.rpt')
        if not os.path.exists(fname):
            url = 'http://www.informatics.jax.org/downloads/reports/' \
                  'MGI_EntrezGene.rpt'
            download_url(url, fname)
        return fname

    def _get_resource_folder(self):
        resource_dir = os.path.join(self.base_folder, 'resources')
        if not os.path.isdir(resource_dir):
            try:
                os.makedirs(resource_dir)
            except Exception:
                logger.warning(resource_dir + ' already exists')
        return resource_dir

    def _replace_outdated_hgnc_symbols(self,pc_old,pc_current):
        logger.info('Replacing outdated HGNC symbols in %s and save as %s' % \
                    (pc_old, pc_current))
        pc = pandas.read_csv(pc_old,sep='\t',dtype=str, header=None)
        col_mapper = {}
        col_mapper[0] = 'source'
        col_mapper[1] = 'rel_type'
        col_mapper[2] = 'target'
        pc = pc.rename(mapper=col_mapper, axis='columns')
        all_symbols = set(pc['source']).union(pc['target'])
        symbol_map = {}
        for sym in all_symbols:
            if not sym.startswith('CHEBI:'):
                hgnc_id = hgnc_client.get_current_hgnc_id(sym)
                if not hgnc_id:
                    continue
                elif isinstance(hgnc_id, list):
                    #outdated gene symbol is ambiguous: maps to multiple genes
                    continue
                latest_symbol = hgnc_client.get_hgnc_name(hgnc_id)
                if latest_symbol != sym:
                    symbol_map[sym] = latest_symbol
        if symbol_map:
            pc.replace(symbol_map,inplace=True)         
        pc.to_csv(pc_current, sep='\t', header=False, index=False)
        os.remove(pc_old)
        
    def download_all(self):
        self.get_go_obo()
        self.get_goa_gaf()
        self.get_pc()
        self.get_mgi_entrez()
        

def download_url(url, fname):
    logger.info('Downloading %s into %s' % (url, fname))
    urllib.request.urlretrieve(url, fname)


def download_gz(fname, url):
    logger.info('Downloading %s and extracting into %s' % (url, fname))
    gz_file = fname + '.gz'
    urllib.request.urlretrieve(url, gz_file)
    with gzip.open(gz_file, 'rb') as fin:
        with open(fname, 'wb') as fout:
            shutil.copyfileobj(fin, fout)


if __name__ == '__main__':
    # Download all the resources if this script is run directly
    ResourceManager().download_all()



"""
This module implements a wrapper around DeepWalk as a class. The class
contains a graph used as a basis for running the Deepwalk algorithm. It also
implements a method to run a given number of walks and save the walks as an
attribute of the instance.
"""
import time
import random
import logging
import functools
import networkx as nx
import multiprocessing
from gensim.models import Word2Vec


logger = logging.getLogger('genewalk.deepwalk')

default_walk_length = 10
default_niter = 100


class DeepWalk(object):
    """Perform DeepWalk (node2vec), i.e., unbiased random walk over nodes
    on an undirected networkx MultiGraph.

    Parameters
    ----------
    graph : networkx.MultiGraph
        A networkx multigraph to be used as the basis for DeepWalk.
    walk_length : Optional[int]
        The length of each random walk on the graph. Default: 10
    niter : Optional[int]
        The number of iterations for each node to run (this is multiplied by
        the number of neighbors of the node when determining the overall number
        of walks to start from a given node). Default: 100

    Attributes
    ----------
    walks : list
        A list of walks.
    """
    def __init__(self, graph, walk_length=default_walk_length,
                 niter=default_niter):
        self.graph = graph
        self.walks = []
        self.wl = walk_length
        self.niter = niter
        self.model = None

    def get_walks(self, workers=1):
        """Generates walks (sentences) sampled by an (unbiased) random walk
        over the networkx MultiGraph.

        Parameters
        ----------
        workers : Optional[int]
            The number of workers to use when running random walks. If greater
            than 1, multiprocessing is used to speed up random walk generation.
            Default: 1
        """
        logger.info('Running random walks...')
        self.walks = []
        start = time.time()
        nodes = nx.nodes(self.graph)
        # In case we don't parallelize
        if workers == 1:
            for count, node in enumerate(nodes):
                walks = run_walks_for_node(node, self.graph, self.niter,
                                           self.wl)
                self.walks.extend(walks)
                if (count + 1) % 100 == 0:
                    logger.info('Walks for %d/%d nodes complete in %.2fs' %
                                (count + 1, len(nodes), time.time() - start))
        # In case we parallelize
        else:
            pool = multiprocessing.Pool(workers)
            start_nodes = get_start_nodes(self.graph, self.niter)
            chunk_size = int(float(0.9*len(start_nodes))/workers)
            walk_fun = functools.partial(run_single_walk,
                                         graph=self.graph,
                                         length=self.wl)
            for count, res in enumerate(
                    pool.imap_unordered(walk_fun, start_nodes,
                                        chunksize=chunk_size)):
                self.walks.append(res)
                if (count + 1) % chunk_size == 0:
                    logger.info('%d/%d walks complete in %.2fs' %
                                (count + 1, len(start_nodes),
                                 time.time() - start))
            logger.debug("Closing pool...")
            pool.close()
            logger.debug("Joining pool...")
            pool.join()
            logger.debug("Pool closed and joined.")
        end = time.time()
        logger.info('Running random walks done in %.2fs' % (end - start))

    def word2vec(self, sg=1, size=8, window=1, min_count=1, negative=5,
                 workers=1, sample=0):
        """Set the model based on Word2Vec
        Source: https://radimrehurek.com/gensim/models/word2vec.html

        Note that his function sets the model attribute if the DeepWalk object
        and doesn't return a value.

        Parameters
        ----------
        sg : Optional[int] {1, 0}
            Defines the training algorithm. If 1, skip-gram is employed;
            otherwise, CBOW is used. For GeneWalk this is set to 1.
        size : Optional[int]
            Dimensionality of the node vectors. Default for GeneWalk is 8.
        window : Optional[int]
            a.k.a. context size. Maximum distance between the current and
            predicted word within a sentence. For GeneWalk this is set to 1
            to assess directly connected nodes only.
        min_count : Optional[int]
            Ignores all words with total frequency lower than this. For
            GeneWalk this is set to 0.
        negative : Optional[int]
            If > 0, negative sampling will be used, the int for negative
            specifies how many "noise words" should be drawn (usually between
            5-20). If set to 0, no negative sampling is used.
            Default for GeneWalk is 5.
        workers : Optional[int]
            Use these many worker threads to train the model (=faster training
            with multicore machines).
        sample : Optional[float]
            The threshold for configuring which higher-frequency words are
            randomly downsampled, useful range is (0, 1e-5). parameter t in eq
            5 Mikolov et al. For GeneWalk this is set to 0.
        """
        logger.info('Generating node vectors...')
        start = time.time()
        self.model = Word2Vec(sentences=self.walks, sg=sg, size=size,
                              window=window, min_count=min_count,
                              negative=negative, workers=workers,
                              sample=sample)
        end = time.time()
        logger.info('Generating node vectors done in %.2fs'
                    % (end - start))


def run_single_walk(start_node, graph, length):
    """Run a single random walk on a graph from a given start node.

    Parameters
    ----------
    graph : networks.MultiGraph
        The graph on which the random walk is to be run.
    start_node : str
        The identifier of the node from which the random walk starts.
    length : int
        The length of the random walk.

    Returns
    -------
    list of str
        A path of the given length, with each element corresponding to a node
        along the path.
    """
    path = [start_node]
    for i in range(1, length):
        start_node = random.choice(list(graph[start_node]))
        path.append(start_node)
    return path


def get_start_nodes(graph, niter):
    start_nodes = []
    for node in nx.nodes(graph):
        start_nodes += [node for _ in range(niter * len(graph[node]))]
    return start_nodes


def run_walks_for_node(node, graph, niter, walk_length):
    """Run all random walks starting from a given node.

    Parameters
    ----------
    node : str
        The identifier of the node from which the walks start.
    graph : networks.MultiGraph
        The graph on which the random walks are to be run.
    niter : int
        The number of iterations to run for gene nodes.
    walk_length : int
        The length of the walk.

    Returns
    -------
    list of list of str
        A list of random walks starting from the given node.
    """
    walks = []
    for _ in range(niter * len(graph[node])):
        walk = run_single_walk(node, graph, walk_length)
        walks.append(walk)
    return walks


def run_walks(graph, **kwargs):
    """Run random walks and get node vectors on a given graph.

    Parameters
    ----------
    graph : networkx.MultiGraph
        The graph on which random walks are going to be run and node vectors
        calculated.
    **kwargs
        Key word arguments passed as the arguments of the DeepWalk constructor,
        as well as the get_walks method and the word2vec method. See the
        DeepWalk class documentation for more information on these.

    Returns
    -------
    :py:class:`genewalk.deepwalk.DeepWalk`
        A DeepWalk instance whose walks attribute contains the list
        of random walks produced on the graph.
    """
    dw_args = {'walk_length': kwargs.pop('walk_length', default_walk_length),
               'niter': kwargs.pop('niter', default_niter)}
    DW = DeepWalk(graph, **dw_args)
    DW.get_walks(kwargs.get('workers', 1))
    DW.word2vec(**kwargs)
    return DW



"""This module implements functions related to the construction of a null
distribution for GeneWalk networks."""
import logging
import networkx as nx

logger = logging.getLogger('genewalk.get_null_distributions')


def get_rand_graph(mg):
    """Return a random graph with the same degree distribution as the input.

    Parameters
    ----------
    mg : networkx.MultiGraph
        An input graph based on which a random graph is generated.

    Returns
    -------
    networkx.MultiGraph
        A random graph whose degree distribution matches that of the output.
    """
    # this is not randomized: order is same
    d_seq = sorted([mg.degree(n) for n in mg.nodes()], reverse=True)
    # creates random multigraph with same degree sequence
    rg = nx.configuration_model(d_seq)
    # the node labels are numbers which gives problems in word2vec
    # so adjust 0 to n0
    mapping = {n: 'n%s' % n for n in rg.nodes()}
    rg = nx.relabel_nodes(rg, mapping, copy=False)
    return rg


def get_null_distributions(rg, nv):
    """Return a distribution with similarity values between (random) node
    vectors originating from the input randomized graph.
    """
    srd = []
    # Generate null distributions from random node vectors
    for node in nx.nodes(rg):
        connections = list(rg[node])
        # only get sims for non self connections
        if node in connections:
            connections.remove(node)
        n_con_source = len(connections)
        if n_con_source > 0:
            sim_dist = list(1-nv.distances(node, other_words=connections))
            srd += sim_dist
    return srd



import logging
default_logger_format = ('%(levelname)s: [%(asctime)s] %(name)s'
                         ' - %(message)s')
default_date_format = '%Y-%m-%d %H:%M:%S'

logging.basicConfig(format=default_logger_format, level=logging.INFO,
                    datefmt=default_date_format)


logger = logging.getLogger('genewalk')
# Remove info about missing package from gensim
logging.getLogger('gensim.summarization.textcleaner').setLevel(logging.WARNING)


__version__ = '1.2.1'



import pickle
import logging
import itertools
import pandas as pd
import networkx as nx
from indra.databases import go_client
from goatools.obo_parser import GODag
from genewalk.resources import ResourceManager
from genewalk.get_indra_stmts import get_famplex_links_from_stmts

logger = logging.getLogger('genewalk.nx_mg_assembler')


def load_network(network_type, network_file, genes, resource_manager=None):
    """Return a network assembler of the given type based on a set of genes.

    Parameters
    ----------
    network_type : str
        The type of the network to be constructed.
    network_file : str
        The path to a file containing information to construct the network.
    genes : list
        A list of gene references.
    resource_manager : Optional[:py:class:`genewalk.resources.ResourceManager`]
        A resource manager object which, if specified, is used to get the
        resource files. Otherwise, the default resource manager is used.

    Returns
    -------
    :py:class:`genewalk.nx_mg_assembler.NxMgAssembler`
        An instance of an NxMgAssembler containing the assembled networkx
        MultiGraph as its graph attribute.
    """
    if not resource_manager:
        resource_manager = None
    if network_type == 'pc':
        mg = PcNxMgAssembler(genes, resource_manager=resource_manager)
    elif network_type == 'indra':
        logger.info('Loading %s' % network_file)
        with open(network_file, 'rb') as fh:
            stmts = pickle.load(fh)
        mg = IndraNxMgAssembler(genes, stmts,
                                resource_manager=resource_manager)
    elif network_type == 'edge_list':
        logger.info('Loading user-provided GeneWalk Network from %s.' %
                    network_file)
        mg = UserNxMgAssembler(network_file, gwn_format='el')
    elif network_type == 'sif':
        logger.info('Loading user-provided GeneWalk Network from %s.' %
                    network_file)
        mg = UserNxMgAssembler(network_file, gwn_format='sif')
    else:
        raise ValueError('Unknown network_type: %s' % network_type)
    return mg


class NxMgAssembler(object):
    """Class which assembles a networkx MultiGraph based on a list of genes.

    Parameters
    ----------
    genes : list of dict
        A list of gene references based on which the graph is assembled.

    Attributes
    ----------
    graph : networkx.MultiGraph
        The assembled graph containing links for interactions between genes,
        GO annotations for genes, and the GO ontology.
    """

    def __init__(self, genes, resource_manager=None):
        self.genes = genes
        self.graph = nx.MultiGraph()
        if not resource_manager:
            self.resource_manager = ResourceManager()
        else:
            self.resource_manager = resource_manager
        self.go_dag = GODag(self.resource_manager.get_go_obo())
        self.goa = self._load_goa_gaf()

    def _get_go_terms_for_gene(self, gene):
        # Filter to rows with the given gene's UniProt ID
        if ('UP' not in gene) or ('HGNC_SYMBOL' not in gene):
            return []
        elif gene['HGNC_SYMBOL'] not in self.graph:
            return []
        df = self.goa[self.goa['DB_ID'] == gene['UP']]
        go_ids = sorted(list(set(df['GO_ID'])))
        return go_ids

    def add_go_annotations(self):
        """Add edges between gene nodes and GO nodes based on GO
        annotations."""
        logger.info('Adding GO annotations for genes in graph.')
        for gene in self.genes:
            go_ids = self._get_go_terms_for_gene(gene)
            for go_id in go_ids:
                if go_id in self.go_dag:
                    go_term = self.go_dag[go_id]
                    if go_term.is_obsolete:
                        continue
                    self.graph.add_node(go_term.id,
                                        name=go_term.name,
                                        GO=go_term.id,
                                        domain=go_term.namespace)
                    self.graph.add_edge(gene['HGNC_SYMBOL'], go_term.id,
                                        label='GO:annotation')

    def add_go_ontology(self):
        """Add edges between GO nodes based on the GO ontology."""
        logger.info('Adding GO ontology edges to graph.')
        for go_term in list(self.go_dag.values()):
            if go_term.is_obsolete:
                continue
            self.graph.add_node(go_term.id,
                                name=go_term.name,
                                GO=go_term.id,
                                domain=go_term.namespace)
            for parent_term in go_term.parents:
                if parent_term.is_obsolete:
                    continue
                self.graph.add_node(go_term.id,
                                    name=go_term.name,
                                    GO=go_term.id,
                                    domain=go_term.namespace)
                self.graph.add_edge(go_term.id, parent_term.id,
                                    label='GO:is_a')

    def node2edges(self, node_key):
        """Return the edges corresponding to a node."""
        return self.graph.edges(node_key, keys=True)

    def save_graph(self, fname):
        """Save the file into a GraphML file.

        Parameters
        ----------
        fname : str
            The name of the file to save the graph into.
        """
        nx.write_graphml(self.graph, fname)

    def _load_goa_gaf(self):
        """Load the gene/GO annotations as a pandas data frame."""
        goa_ec = {'EXP', 'IDA', 'IPI', 'IMP', 'IGI', 'IEP', 'HTP', 'HDA',
                  'HMP', 'HGI', 'HEP', 'IBA', 'IBD'}
        goa = pd.read_csv(self.resource_manager.get_goa_gaf(), sep='\t',
                          skiprows=23, dtype=str,
                          header=None,
                          names=['DB',
                                 'DB_ID',
                                 'DB_Symbol',
                                 'Qualifier',
                                 'GO_ID',
                                 'DB_Reference',
                                 'Evidence_Code',
                                 'With_From',
                                 'Aspect',
                                 'DB_Object_Name',
                                 'DB_Object_Synonym',
                                 'DB_Object_Type',
                                 'Taxon',
                                 'Date',
                                 'Assigned',
                                 'Annotation_Extension',
                                 'Gene_Product_Form_ID'])
        goa = goa.sort_values(by=['DB_ID', 'GO_ID'])
        # Filter out all "NOT" negative evidences
        goa['Qualifier'].fillna('', inplace=True)
        goa = goa[~goa['Qualifier'].str.startswith('NOT')]
        # Filter to rows with evidence code corresponding to experimental
        # evidence
        goa = goa[goa['Evidence_Code'].isin(goa_ec)]
        return goa


class PcNxMgAssembler(NxMgAssembler):
    """The PcNxMgAssembler assembles a GeneWalk Network with gene reactions
    from Pathway Commons and GO ontology and annotations into a networkx
    (undirected)  MultiGraph including edge attributes.

    Parameters
    ----------
    genes : list fo dict
        A list of gene references based on which the network is assembled.

    Attributes
    ----------
    graph : networkx.MultiGraph
        A GeneWalk Network that is assembled by this assembler.
    """
    def __init__(self, genes, resource_manager=None):
        super().__init__(genes, resource_manager)
        self.add_pc_edges()
        self.add_go_annotations()
        self.add_go_ontology()

    def add_pc_edges(self):
        """Add edges between gene nodes based on PathwayCommons
        interactions."""
        logger.info('Adding gene edges from Pathway Commons to graph.')
        gwn_df = pd.read_csv(self.resource_manager.get_pc(), sep='\t',
                             dtype=str, header=None)
        col_mapper = {}
        col_mapper[0] = 'source'
        col_mapper[1] = 'rel_type'
        col_mapper[2] = 'target'
        edge_attributes = True
        gwn_df = gwn_df.rename(mapper=col_mapper, axis='columns')
        pc = nx.from_pandas_edgelist(gwn_df, source='source', target='target',
                                     edge_attr=edge_attributes,
                                     create_using=nx.MultiGraph)
        # subset over genes in the input gene list
        hgnc_symbols = [g['HGNC_SYMBOL'] for g in self.genes]
        hgnc_ids = [g['HGNC'] for g in self.genes]
        up_ids = [g['UP'] for g in self.genes]
        pc.add_nodes_from(hgnc_symbols)
        pc_sub = pc.subgraph(hgnc_symbols)
        gene2hgnc_dict = dict(zip(hgnc_symbols, hgnc_ids))
        nx.set_node_attributes(pc_sub, gene2hgnc_dict, 'HGNC')
        gene2up_dict = dict(zip(hgnc_symbols, up_ids))
        nx.set_node_attributes(pc_sub, gene2up_dict, 'UP')
        # Make a copy to unfreeze graph
        self.graph = nx.MultiGraph(pc_sub)
        logger.info('Number of PC originating nodes %d' %
                    nx.number_of_nodes(self.graph))


class IndraNxMgAssembler(NxMgAssembler):
    """The IndraNxMgAssembler assembles INDRA Statements and GO ontology /
    annotations into a networkx (undirected) MultiGraph including edge
    attributes. This code is based on INDRA's SifAssembler
    http://indra.readthedocs.io/en/latest/_modules/indra/assemblers/sif_assembler.html

    Parameters
    ----------
    stmts : list[indra.statements.Statement]
        A list of INDRA Statements to be added to the assembler's list
        of Statements.

    Attributes
    ----------
    graph : networkx.MultiGraph
        A GeneWalk Network that is assembled by this assembler.
    """
    def __init__(self, genes, stmts, resource_manager=None):
        self.indra_nodes = set()
        self.stmts = stmts
        super().__init__(genes, resource_manager)
        self.add_indra_edges()
        self.add_fplx_edges()
        self.add_go_annotations()
        self.add_go_ontology()

    def add_indra_edges(self):
        """Add edges between gene nodes and GO nodes based on INDRA Statements.
        """
        logger.info('Adding nodes from INDRA statements.')
        for i, st in enumerate(self.stmts):
            # Get all agents in the statement
            agents = [a for a in st.agent_list() if a is not None]
            # Only include edges for statements with at least 2 Agents
            # excludes (irrelevant) stmt types: Translocation, ActiveForm,
            # SelfModification
            if len(agents) < 2:
                continue
            # Create a label that is unique to the statement and its type
            edge_type = type(st).__name__
            edge_key = '%d_%s' % (i, edge_type)
            # Iterate over all the agent combinations and add edge
            for a, b in itertools.combinations(agents, 2):
                a_node = self.add_agent_node(a)
                b_node = self.add_agent_node(b)
                self.graph.add_edge(a_node, b_node, key=edge_key,
                                    label=edge_type)
        hgnc_symbols = [g['HGNC_SYMBOL'] for g in self.genes]
        hgnc_ids = [g['HGNC'] for g in self.genes]
        up_ids = [g['UP'] for g in self.genes]
        self.graph.add_nodes_from(hgnc_symbols)
        gene2hgnc_dict = dict(zip(hgnc_symbols, hgnc_ids))
        nx.set_node_attributes(self.graph, gene2hgnc_dict, 'HGNC')
        gene2up_dict = dict(zip(hgnc_symbols, up_ids))
        nx.set_node_attributes(self.graph, gene2up_dict, 'UP')
        logger.info('Number of INDRA originating nodes %d.' %
                    len(self.indra_nodes))

    def add_fplx_edges(self):
        """Add edges between gene nodes and families/complexes they are part
        of."""
        links = get_famplex_links_from_stmts(self.stmts)
        for s, t in links:
            self.graph.add_edge(s, t, label='FPLX:is_a')

    def add_agent_node(self, agent):
        """Add a node corresponding to an INDRA Agent."""
        go_id = agent.db_refs.get('GO')
        if go_id:
            go_id = go_id if go_id.startswith('GO:') else 'GO:%s' % go_id
            node_key = go_id
            name = go_client.get_go_label(go_id)
            self.graph.add_node(node_key, name=name,
                                source='indra', **agent.db_refs)
        else:
            node_key = agent.name
            self.graph.add_node(node_key, name=agent.name, **agent.db_refs,
                                source='indra')
        self.indra_nodes.add(node_key)
        return node_key

    def node2stmts(self, node_key):
        """Return the INDRA Statements given the key of a graph node."""
        matching_stmts = []
        node_name = self.graph.nodes[node_key]['name']
        for stmt in self.stmts:
            for agent in stmt.agent_list():
                if agent is not None:
                    agent_name = agent.name
                    if agent_name == node_name:
                        matching_stmts.append(stmt)
                        break
        return matching_stmts


class UserNxMgAssembler(object):
    """Loads a user-provided GeneWalk Network from a given file.

    Parameters
    ----------
    filepath : str
        Path to the user-provided genewalk network file, assumed to contain
        gene symbols and GO IDs. See gwn_format for supported format details.
    gwn_format : Optional[str]
        'el' (default, edge list: nodeA nodeB (if more columns
        present: interpreted as edge attributes) \
        or 'sif' (simple interaction format: nodeA <relationship type> nodeB).
        Do not include column headers.

    Attributes
    ----------
    graph : networkx.MultiGraph
        A GeneWalk Network that is loaded by this assembler.
    """
    def __init__(self, filepath, gwn_format='el'):
        self.graph = nx.MultiGraph()
        self.filepath = filepath
        self.gwn_format = gwn_format
        self.add_network_edges()

    def add_network_edges(self):
        """Assemble the GeneWalk Network from the user-provided file path."""
        gwn_df = pd.read_csv(self.filepath, dtype=str, header=None)
        col_mapper = {}
        if self.gwn_format == 'el':
            col_mapper[0] = 'source'
            col_mapper[1] = 'target'
            if len(gwn_df.columns) > 2:
                col_mapper[2] = 'rel_type'
                if len(gwn_df.columns) > 3:
                    for c in gwn_df.columns[3:]:
                        col_mapper[c] = 'edge_attr'+str(c-1)
                edge_attributes = True
            else:
                edge_attributes = False
        elif self.gwn_format == 'sif':
            col_mapper[0] = 'source'
            col_mapper[1] = 'rel_type'
            col_mapper[2] = 'target'
            edge_attributes = True

        gwn_df.rename(mapper=col_mapper, axis='columns')
        self.graph = nx.from_pandas_edgelist(gwn_df, 'source', 'target',
                                             edge_attr=edge_attributes,
                                             create_using=nx.MultiGraph)



"""This script creates a pickle file of INDRA Statements for a specific set of
genes, based on a pre-assembled network of interactions produced by INDRA. It
loads the INDRA interactions as a pandas DataFrame and, filters it to the genes
and familiex/complexes of interest, as well as targeted biological processes.
It downloads the relevant Statement objects for reference and dumps them into
a pickle file.
"""
import pandas
import pickle
import logging
import argparse
from indra.util import batch_iter
from indra.databases import go_client
from indra.sources import indra_db_rest
from indra.databases import hgnc_client
from indra.preassembler.hierarchy_manager import hierarchies


logger = logging.getLogger('genewalk.get_indra_stmts')


def load_genes(fname):
    """Return a list of genes IDs from a file with lines like HGNC:123."""
    with open(fname, 'r') as fh:
        # Get the HGNC IDs from the list of genes file, assuming that
        # each line looks like HGNC:123
        genes = [l.strip().split(':')[1] for l in fh.readlines()]
    logger.info('Loaded %d genes from %s' % (len(genes), fname))
    return genes


def load_mouse_genes(fname):
    """Return a list of human genes based on a table of mouse genes."""
    # assumes the csv has headers
    df = pandas.read_csv(fname)
    for c in df.columns:
        # assumes the first column starting with MGI is the relevant one
        # with MGI:IDs
        if c.startswith('MGI'):
            df = df.rename(columns={c: 'MGI'})
            break
    mgi_ids = df['MGI']
    genes = []
    for mgi_id in mgi_ids:
        if mgi_id.startswith('MGI:'):
            mgi_id = mgi_id[4:]
        hgnc_id = hgnc_client.get_hgnc_from_mouse(mgi_id)
        if not hgnc_id:
            print('Could not find human gene corresponding to MGI %s' % mgi_id)
            continue
        genes.append(hgnc_id)
    return genes


def load_indra_df(fname):
    """Return an INDRA Statement data frame from a pickle file."""
    with open(fname, 'rb') as fh:
        df = pickle.load(fh)
    logger.info('Loaded %d rows from %s' % (len(df), fname))
    return df


def dump_pickle(stmts, fname):
    """Dump a list of Statements into a picke file."""
    with open(fname, 'wb') as fh:
        pickle.dump(stmts, fh)
    logger.info('Dumped %d statements into %s' % (len(stmts), fname))


def filter_to_genes(df, genes, fplx_terms):
    """Filter a data frame of INDRA Statements given gene and FamPlex IDs."""
    # Look for sources that are in the gene list or whose families/complexes
    # are in the FamPlex term list
    source_filter = (((df.agA_ns == 'HGNC') & (df.agA_id.isin(genes))) |
                     ((df.agA_ns == 'FPLX') & (df.agA_id.isin(fplx_terms))))
    # Look for targets that are in the gene list or whose families/complexes
    # are in the FamPlex term list, or which are GO terms
    target_filter = (((df.agB_ns == 'HGNC') & (df.agB_id.isin(genes))) |
                     ((df.agB_ns == 'FPLX') & (df.agB_id.isin(fplx_terms))) |
                     (df.agB_ns == 'GO'))
    # sources/targets
    df = df[source_filter & target_filter]
    logger.info('Filtered data frame to %d rows.' % len(df))
    return df


def get_gene_parents(hgnc_id):
    eh = hierarchies['entity']
    gene_uri = eh.get_uri('HGNC', hgnc_id)
    parents = eh.get_parents(gene_uri)
    parent_ids = [eh.ns_id_from_uri(par_uri)[1] for par_uri in parents]
    return parent_ids


def get_famplex_terms(genes):
    """Get a list of associated FamPlex IDs from a list of gene IDs."""
    all_parents = set()
    for hgnc_id in genes:
        parent_ids = get_gene_parents(hgnc_id)
        all_parents |= set(parent_ids)
    fplx_terms = sorted(list(all_parents))
    logger.info('Found %d relevant FamPlex terms.' % (len(fplx_terms)))
    return fplx_terms


def get_famplex_links(df, fname):
    """Given a list of INDRA Statements, construct FamPlex links."""
    genes_appearing = (set(df[df.agA_ns == 'HGNC'].agA_name) |
                       set(df[df.agB_ns == 'HGNC'].agB_name))
    fplx_appearing = (set(df[df.agA_ns == 'FPLX'].agA_id) |
                      set(df[df.agB_ns == 'FPLX'].agB_id))
    links = get_famplex_links_from_lists(genes_appearing, fplx_appearing)
    with open(fname, 'w') as fh:
        for link in links:
            fh.write('%s,%s\n' % link)


def get_famplex_links_from_stmts(stmts):
    genes_appearing = set()
    fplx_appearing = set()
    for stmt in stmts:
        agents = [a for a in stmt.agent_list() if a is not None]
        if len(agents) < 2:
            continue
        for agent in agents:
            if 'HGNC' in agent.db_refs:
                genes_appearing.add(agent.name)
            elif 'FPLX' in agent.db_refs:
                fplx_appearing.add(agent.name)
    return get_famplex_links_from_lists(genes_appearing, fplx_appearing)


def get_famplex_links_from_lists(genes_appearing, fplx_appearing):
    links = []
    for gene in genes_appearing:
        parent_ids = get_gene_parents(gene)
        parents_appearing = fplx_appearing & set(parent_ids)
        links += [(gene, parent) for parent in parents_appearing]
    eh = hierarchies['entity']
    for fplx_child in fplx_appearing:
        fplx_uri = eh.get_uri('FPLX', fplx_child)
        parents = eh.get_parents(fplx_uri)
        parent_ids = [eh.ns_id_from_uri(par_uri)[1] for par_uri in parents]
        parents_appearing = fplx_appearing & set(parent_ids)
        links += [(fplx_child, parent) for parent in parents_appearing]
    return links


def download_statements(df, ev_limit=5):
    """Download the INDRA Statements corresponding to entries in a data frame.
    """
    all_stmts = []
    for idx, group in enumerate(batch_iter(df.hash, 500)):
        logger.info('Getting statement batch %d' % idx)
        idbp = indra_db_rest.get_statements_by_hash(list(group),
                                                    ev_limit=ev_limit)
        all_stmts += idbp.statements
    return all_stmts


def remap_go_ids(stmts):
    for stmt in stmts:
        for agent in stmt.agent_list():
            if agent is not None and 'GO' in agent.db_refs:
                prim_id = go_client.get_primary_id(agent.db_refs['GO'])
                if prim_id:
                    agent.db_refs['GO'] = prim_id


if __name__ == '__main__':
    # Handle command line arguments
    # TODO: make specific sets of statements for the use cases available in
    #  the repo (or on S3) and parameterize here to be able to load them.
    parser = argparse.ArgumentParser(
        description='Choose a file with a list of genes to get a SIF for.')
    parser.add_argument('--df', default='data/stmt_df.pkl')
    parser.add_argument('--genes', default='data/JQ1_HGNCidForINDRA.csv')
    parser.add_argument('--mouse_genes')
    parser.add_argument('--stmts', default='data/JQ1_HGNCidForINDRA_stmts.pkl')
    args = parser.parse_args()

    # Load genes and get FamPlex terms
    if args.mouse_genes:
        genes = load_mouse_genes(args.mouse_genes)
    else:
        genes = load_genes(args.genes)
    fplx_terms = get_famplex_terms(genes)
    # Load INDRA Statements in a flat data frame
    df = load_indra_df(args.df)
    # Filter the data frame to relevant entities
    df = filter_to_genes(df, genes, fplx_terms)
    # Download the Statement corresponding to each row
    stmts = download_statements(df)
    # Remap any outdated GO IDs
    remap_go_ids(stmts)
    # Dump the Statements into a pickle file
    dump_pickle(stmts, args.stmts)



import gc
import os
import sys
import copy
import pickle
import random
import logging
import argparse
import numpy as np
from genewalk import __version__
from genewalk.nx_mg_assembler import load_network
from genewalk.gene_lists import read_gene_list
from genewalk.deepwalk import run_walks
from genewalk.null_distributions import get_rand_graph, \
    get_null_distributions
from genewalk.perform_statistics import GeneWalk
from genewalk import logger as root_logger, default_logger_format, \
    default_date_format
from genewalk.resources import ResourceManager

logger = logging.getLogger('genewalk.cli')

default_base_folder = os.path.join(os.path.expanduser('~/'), 'genewalk')


def create_project_folder(base_folder, project):
    project_folder = os.path.join(base_folder, project)
    logger.info('Creating project folder at %s' % project_folder)
    if not os.path.exists(project_folder):
        os.makedirs(project_folder)
    return project_folder


def save_pickle(obj, project_folder, prefix):
    fname = os.path.join(project_folder, '%s.pkl' % prefix)
    logger.info('Saving into %s...' % fname)
    with open(fname, 'wb') as fh:
        pickle.dump(obj, fh)


def load_pickle(project_folder, prefix):
    fname = os.path.join(project_folder, '%s.pkl' % prefix)
    logger.info('Loading %s...' % fname)
    with open(fname, 'rb') as fh:
        return pickle.load(fh)


def main():
    parser = argparse.ArgumentParser(
        description='Run GeneWalk on a list of genes provided in a text '
                    'file.')
    parser.add_argument('--version', action='version',
                        version='GeneWalk %s' % __version__,
                        help='Print the version of GeneWalk and exit.')
    parser.add_argument('--project', help='A name for the project which '
                                          'determines the folder within the '
                                          'base folder in which the '
                                          'intermediate and final results '
                                          'are written. Must contain only '
                                          'characters that are valid in '
                                          'folder names.',
                        required=True)
    parser.add_argument('--genes', help='Path to a text file with a list of '
                                        'genes of interest, for example'
                                        'differentially expressed genes. '
                                        'The type of gene identifiers used in '
                                        'the text file are provided in the '
                                        'id_type argument.',
                        required=True)
    parser.add_argument('--id_type',
                        help='The type of gene IDs provided in the text file '
                             'in the genes argument. Possible values are: '
                             'hgnc_symbol, hgnc_id, ensembl_id, and mgi_id.',
                        choices=['hgnc_symbol', 'hgnc_id',
                                 'ensembl_id', 'mgi_id', 'entrez_human',
                                 'entrez_mouse'],
                        required=True)
    parser.add_argument('--stage', default='all',
                        help='The stage of processing to run. Default: '
                             '%(default)s',
                        choices=['all', 'node_vectors', 'null_distribution',
                                 'statistics'])
    parser.add_argument('--base_folder', default=default_base_folder,
                        help='The base folder used to store GeneWalk '
                             'temporary and result files for a given project.'
                             ' Default: %(default)s')
    parser.add_argument('--network_source', default='pc',
                        help='The source of the network to be used.'
                             'Possible values are: pc, indra, edge_list, and '
                             'sif. In case of indra, edge_list, and sif, '
                             'the network_file argument must be specified.'
                             ' Default: %(default)s',
                        choices=['pc', 'indra', 'edge_list', 'sif'])
    parser.add_argument('--network_file', default=None,
                        help='If network_source is indra, this argument '
                             'points to a Python pickle file in which a list '
                             'of INDRA Statements constituting the network '
                             'is contained. In case network_source is '
                             'edge_list or sif, '
                             'the network_file argument points to a text file '
                             'representing the network.')
    parser.add_argument('--nproc', default=1, type=int,
                        help='The number of processors to use in a '
                             'multiprocessing environment. Default: '
                             '%(default)s')
    parser.add_argument('--nreps_graph', default=3, type=int,
                        help='The number of repeats to run when calculating '
                             'node vectors on the GeneWalk graph. '
                             'Default: %(default)s')
    parser.add_argument('--nreps_null', default=3, type=int,
                        help='The number of repeats to run when calculating '
                             'node vectors on the random network graphs '
                             'for constructing the null distribution. '
                             'Default: %(default)s')
    parser.add_argument('--alpha_fdr', default=1, type=float,
                        help='The false discovery rate to use when '
                             'outputting the final statistics table. '
                             'If 1 (default), all similarities are output, '
                             'otherwise only the ones whose false discovery '
                             'rate are below this parameter are included. '
                             'Default: %(default)s')
    parser.add_argument('--save_dw', default=False, type=bool,
                        help='If True, the full DeepWalk object for each '
                             'repeat is saved in the project folder. This can '
                             'be useful for debugging but the files are '
                             'typically very large. Default: %(default)s')
    parser.add_argument('--random_seed', default=None, type=int,
                        help='If provided, the random number generator is '
                             'seeded with the given value. This should only '
                             'be used if the goal is to deterministically '
                             'reproduce a prior result obtained with the same '
                             'random seed.')
    args = parser.parse_args()

    # Now we run the relevant stage of processing
    project_folder = create_project_folder(args.base_folder, args.project)

    # Add a logger specific to the project and processing stage
    log_file = os.path.join(project_folder, 'genewalk_%s.log' % args.stage)
    formatter = logging.Formatter(default_logger_format,
                                  datefmt=default_date_format)
    project_log_handler = logging.FileHandler(log_file)
    project_log_handler.setFormatter(formatter)
    root_logger.addHandler(project_log_handler)

    if args.random_seed:
        logger.info('Running with random seed %d' % args.random_seed)
        random.seed(a=int(args.random_seed))

    # Make sure we have all the resource files
    rm = ResourceManager(base_folder=args.base_folder)
    rm.download_all()

    if args.stage in ('all', 'node_vectors'):
        genes = read_gene_list(args.genes, args.id_type, rm)
        save_pickle(genes, project_folder, 'genes')
        MG = load_network(args.network_source, args.network_file, genes,
                          resource_manager=rm)
        save_pickle(MG.graph, project_folder, 'multi_graph')
        for i in range(args.nreps_graph):
            logger.info('%s/%s' % (i + 1, args.nreps_graph))
            DW = run_walks(MG.graph, workers=args.nproc)

            # Pickle the node vectors (embeddings) and DW object
            if args.save_dw:
                save_pickle(DW, project_folder, 'deepwalk_%d' % (i + 1))
            nv = copy.deepcopy(DW.model.wv)
            save_pickle(nv, project_folder,
                        'deepwalk_node_vectors_%d' % (i + 1))

            # Delete the DeepWalk object to clear memory
            del DW, nv
            gc.collect()

    if args.stage in ('all', 'null_distribution'):
        MG = load_pickle(project_folder, 'multi_graph')
        srd = []
        for i in range(args.nreps_null):
            logger.info('%s/%s' % (i + 1, args.nreps_null))
            RG = get_rand_graph(MG)
            DW = run_walks(RG, workers=args.nproc)

            # Pickle the node vectors (embeddings) and DW object
            if args.save_dw:
                save_pickle(DW, project_folder, 'deepwalk_rand_%d' % (i + 1))
            nv = copy.deepcopy(DW.model.wv)
            save_pickle(nv, project_folder, 'deepwalk_node_vectors_rand_%d'
                                            % (i + 1))
            # Delete the DeepWalk object to clear memory
            del DW
            gc.collect()

            # Calculate the null distributions
            srd += get_null_distributions(RG, nv)
            del nv
            gc.collect()
        srd = np.asarray(sorted(srd))
        save_pickle(srd, project_folder, 'genewalk_rand_simdists')

    if args.stage in ('all', 'statistics'):
        MG = load_pickle(project_folder, 'multi_graph')
        genes = load_pickle(project_folder, 'genes')
        nvs = [load_pickle(project_folder,
                           'deepwalk_node_vectors_%d' % (i + 1))
               for i in range(args.nreps_graph)]
        null_dist = load_pickle(project_folder, 'genewalk_rand_simdists')
        GW = GeneWalk(MG, genes, nvs, null_dist)
        df = GW.generate_output(alpha_fdr=args.alpha_fdr,
                                base_id_type=args.id_type)
        fname = os.path.join(project_folder, 'genewalk_results.csv')
        logger.info('Saving final results into %s' % fname)
        df.to_csv(fname, index=False, float_format='%.3e')


if __name__ == '__main__':
    main()



import csv
import pickle
import logging
from io import StringIO
from os.path import join, abspath, dirname
from collections import Counter
import requests
from indra.databases import hgnc_client, uniprot_client


logger = logging.getLogger('genewalk.gene_lists')


def read_gene_list(fname, id_type, resource_manager):
    """Return references for genes from a file with the given ID type.

    Parameters
    ----------
    fname : str
        The name of the file containing the list of genes. Each line of the
        file corresponds to a single gene.
    id_type : str
        The type of identifier contained in each line of the gene list file.
        Possible values are: hgnc_symbol, hgnc_id, ensembl_id, mgi_id.
    resource_manager : genewalk.resources.ResourceManager
        ResourceManager object, used to obtain entrez-mgi mappings if
        necessary.

    Returns
    -------
    dict
        A dictionary of references with keys including HGNCSYMBOL, HGNC, UP,
        and if id_type is mgi_id, MGI, with values corresponding to the
        identifiers of the provided list of genes.
    """
    with open(fname, 'r') as fh:
        # This is to make the list unique while preserving
        # the original order as much as possible
        unique_lines = []
        for line in fh.readlines():
            line = line.strip()
            if line not in unique_lines:
                unique_lines.append(line)
    if id_type == 'hgnc_symbol':
        refs = map_hgnc_symbols(unique_lines)
    elif id_type == 'hgnc_id':
        refs = map_hgnc_ids(unique_lines)
    elif id_type == 'ensembl_id':
        refs = map_ensembl_ids(unique_lines)
    elif id_type == 'mgi_id':
        refs = map_mgi_ids(unique_lines)
    elif id_type == 'entrez_human':
        refs = map_entrez_human(unique_lines)
    elif id_type == 'entrez_mouse':
        refs = map_entrez_mouse(unique_lines, resource_manager)
    else:
        raise ValueError('Unknown id_type: %s' % id_type)
    if not refs:
        raise ValueError('None of the IDs in %s could be mapped. It is '
                         'likely that the file uses an ID type or format '
                         'that GeneWalk cannot interpret.' % fname)
    return refs


def map_hgnc_symbols(hgnc_symbols):
    """Return references based on a list of HGNC symbols."""
    refs = []
    for hgnc_symbol in hgnc_symbols:
        ref = {'HGNC_SYMBOL': hgnc_symbol, 'HGNC': None, 'UP': None}
        hgnc_id = hgnc_client.get_current_hgnc_id(hgnc_symbol)
        if not hgnc_id:
            logger.warning('Could not get HGNC ID for symbol %s' % hgnc_symbol)
            continue
        elif isinstance(hgnc_id, list):
            logger.warning('More than one current HGNC ID for outdated '
                           'symbol %s' % hgnc_symbol)
            continue
        ref['HGNC'] = hgnc_id
        uniprot_id = hgnc_client.get_uniprot_id(hgnc_id)
        if not uniprot_id:
            logger.warning('Could not get UniProt ID for symbol %s' %
                           hgnc_symbol)
            continue
        ref['UP'] = uniprot_id
        refs.append(ref)
    return refs


def map_hgnc_ids(hgnc_ids):
    """Return references based on a list of HGNC IDs."""
    refs = []
    for hgnc_id in hgnc_ids:
        if hgnc_id.startswith('HGNC:'):
            hgnc_id = hgnc_id[5:]
        hgnc_ref = _refs_from_hgnc_id(hgnc_id)
        if hgnc_ref is None:
            continue
        refs.append(hgnc_ref)
    return refs


def _refs_from_hgnc_id(hgnc_id):
    ref = {'HGNC_SYMBOL': None, 'HGNC': hgnc_id, 'UP': None}
    hgnc_name = hgnc_client.get_hgnc_name(hgnc_id)
    if not hgnc_name:
        logger.warning('Could not get HGNC name for ID %s' %
                       hgnc_id)
        return None
    ref['HGNC_SYMBOL'] = hgnc_name
    uniprot_id = hgnc_client.get_uniprot_id(hgnc_id)
    if not uniprot_id:
        logger.warning('Could not get UniProt ID for HGNC ID %s' %
                       hgnc_id)
        return None
    ref['UP'] = uniprot_id
    return ref


def map_mgi_ids(mgi_ids):
    """Return references based on a list of MGI IDs."""
    refs = []
    for mgi_id in mgi_ids:
        if mgi_id.startswith('MGI:'):
            mgi_id = mgi_id[4:]
        mgi_ref = _refs_from_mgi_id(mgi_id)
        if mgi_ref is None:
            continue
        refs.append(mgi_ref)
    return refs


def _refs_from_mgi_id(mgi_id):
    ref = {'MGI': mgi_id}
    hgnc_id = hgnc_client.get_hgnc_from_mouse(mgi_id)
    if hgnc_id is None:
        logger.warning('Could not get HGNC ID for MGI ID %s' %
                       mgi_id)
        return None
    hgnc_ref = _refs_from_hgnc_id(hgnc_id)
    if hgnc_ref is None:
        return None
    ref.update(hgnc_ref)
    return ref


def map_ensembl_ids(ensembl_ids):
    """Return references based on a list of Ensembl IDs."""
    refs = []
    for ensembl_id in ensembl_ids:
        ref = {'ENSEMBL': ensembl_id}
        ensembl_id = ensembl_id.split('.', maxsplit=1)[0]
        hgnc_id = hgnc_client.get_hgnc_from_ensembl(ensembl_id)
        if not hgnc_id:
            logger.warning('Could not get HGNC ID for ENSEMBL ID %s' %
                           ensembl_id)
            continue
        hgnc_ref = _refs_from_hgnc_id(hgnc_id)
        if hgnc_ref is None:
            continue
        ref.update(hgnc_ref)
        refs.append(ref)
    return refs


def map_entrez_human(entrez_ids):
    """Return references based on human Entrez gene IDs."""
    refs = []
    for entrez_id in entrez_ids:
        ref = {'EGID': entrez_id}
        hgnc_id = hgnc_client.get_hgnc_from_entrez(entrez_id)
        if hgnc_id is None:
            logger.warning("Could not find HGNC ID for Entrez ID %s" %
                           entrez_id)
            continue
        hgnc_ref = _refs_from_hgnc_id(hgnc_id)
        if hgnc_ref is None:
            continue
        ref.update(hgnc_ref)
        refs.append(ref)
    return refs


def map_entrez_mouse(entrez_ids, rm):
    """Return references based on mouse Entrez gene IDs."""
    # Get the entrez file path from the resource manager
    mgi_entrez_file = rm.get_mgi_entrez()
    # Process the MGI-Entrez mapping file
    entrez_to_mgi = {}
    with open(mgi_entrez_file, 'r') as f:
        csvreader = csv.reader(f, delimiter='\t')
        for row in csvreader:
            # Remove "MGI:" prefix
            mgi = row[0][4:]
            entrez = row[8]
            entrez_to_mgi[entrez] = mgi
    refs = []
    for entrez_id in entrez_ids:
        mgi_id = entrez_to_mgi.get(entrez_id)
        if not mgi_id:
            logger.warning("Could not find an MGI mapping for Entrez ID %s"
                           % entrez_id)
            continue
        ref = {'EGID': entrez_id, 'MGI': mgi_id}
        mgi_refs = _refs_from_mgi_id(mgi_id)
        if mgi_refs is None:
            continue
        ref.update(mgi_refs)
        refs.append(ref)
    return refs



from networkx import MultiGraph
from genewalk.deepwalk import DeepWalk


def test_run_walks():
    mg = MultiGraph()
    mg.add_edge('a', 'b')
    mg.add_edge('b', 'c')
    dw = DeepWalk(mg, 10, 100)
    dw.get_walks(workers=1)
    # Number of neighbors for all nodes together is 4, times niter: 400
    assert len(dw.walks) == 400, len(dw.walks)
    assert len([w for w in dw.walks if w[0] == 'a']) == 100
    assert len([w for w in dw.walks if w[0] == 'b']) == 200

    dw.get_walks(workers=2)
    # Number of neighbors for all nodes together is 4, times niter: 400
    assert len(dw.walks) == 400, len(dw.walks)
    assert len([w for w in dw.walks if w[0] == 'a']) == 100
    assert len([w for w in dw.walks if w[0] == 'b']) == 200



from nose.tools import raises
from genewalk.gene_lists import *
from genewalk.cli import default_base_folder
from genewalk.resources import ResourceManager

def test_map_lists():
    refs = map_hgnc_symbols(['BRAF', 'KRAS'])
    assert refs[0]['HGNC'] == '1097', refs
    assert refs[0]['UP'] == 'P15056', refs
    assert refs[0]['HGNC_SYMBOL'] == 'BRAF', refs
    assert refs[1]['HGNC'] == '6407', refs
    assert refs[1]['UP'] == 'P01116', refs
    assert refs[1]['HGNC_SYMBOL'] == 'KRAS', refs

    refs = map_hgnc_ids(['1097', '6407'])
    assert refs[0]['HGNC'] == '1097', refs
    assert refs[0]['UP'] == 'P15056', refs
    assert refs[0]['HGNC_SYMBOL'] == 'BRAF', refs
    assert refs[1]['HGNC'] == '6407', refs
    assert refs[1]['UP'] == 'P01116', refs
    assert refs[1]['HGNC_SYMBOL'] == 'KRAS', refs

    refs = map_mgi_ids(['MGI:892970'])
    assert refs[0]['HGNC'] == '6817', refs
    assert refs[0]['HGNC_SYMBOL'] == 'MAL', refs
    assert refs[0]['UP'] == 'P21145', refs

    refs = map_ensembl_ids(['ENSG00000157764'])
    assert refs[0]['HGNC'] == '1097', refs
    assert refs[0]['UP'] == 'P15056', refs
    assert refs[0]['HGNC_SYMBOL'] == 'BRAF', refs

    refs = map_ensembl_ids(['ENSG00000157764.9'])
    assert refs[0]['HGNC'] == '1097', refs
    assert refs[0]['UP'] == 'P15056', refs
    assert refs[0]['HGNC_SYMBOL'] == 'BRAF', refs


def test_read_gene_list():
    with open('test_gene_list.txt', 'w') as fh:
        fh.write('HGNC:1097')
    refs = read_gene_list('test_gene_list.txt', 'hgnc_id', None)
    assert len(refs) == 1


@raises(ValueError)
def test_read_gene_list_bad():
    with open('test_gene_list.txt', 'w') as fh:
        fh.write('HGNC:1097')
    refs = read_gene_list('test_gene_list.txt', 'ensembl_id', None)
    assert len(refs) == 1


def test_read_gene_list_entrez_human():
    with open('test_gene_list_entrez_human.txt', 'w') as fh:
        fh.write('2597')
    refs = read_gene_list('test_gene_list_entrez_human.txt',
                          'entrez_human', None)
    assert refs[0]['HGNC_SYMBOL'] == 'GAPDH'
    assert len(refs) == 1


def test_read_gene_list_entrez_mouse():
    rm = ResourceManager(base_folder=default_base_folder)
    with open('test_gene_list_entrez_mouse.txt', 'w') as fh:
        fh.write('14433')
    refs = read_gene_list('test_gene_list_entrez_mouse.txt',
                          'entrez_mouse', rm)
    assert len(refs) == 1
    assert refs[0]['MGI'] == '95640'
    assert refs[0]['HGNC_SYMBOL'] == 'GAPDH'




from os import path
from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


def main():
    install_list = ['numpy', 'pandas', 'networkx>=2.1', 'gensim', 'goatools',
                    'indra>=1.14.1', 'scipy>=1.3.0']

    setup(name='genewalk',
          version='1.2.1',
          description='Determine gene function based on network embeddings.',
          long_description=long_description,
          long_description_content_type='text/markdown',
          author='Robert Ietswaart',
          author_email='robert_ietswaart@hms.harvard.edu',
          url='https://github.com/churchmanlab/genewalk',
          classifiers=[
            'Development Status :: 4 - Beta',
            'Environment :: Console',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: BSD License',
            'Operating System :: OS Independent',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Topic :: Scientific/Engineering :: Bio-Informatics',
            ],
          keywords=['gene function', 'network', 'embedding'],
          packages=find_packages(),
          install_requires=install_list,
          tests_require=['nose'],
          include_package_data=True,
          entry_points={'console_scripts': ['genewalk = genewalk.cli:main']},
        )


if __name__ == '__main__':
    main()



import logging
import pandas as pd
import numpy as np
import networkx as nx
from statsmodels.stats.multitest import fdrcorrection
from scipy.stats import gmean, gstd

logger = logging.getLogger('genewalk.perform_statistics')


class GeneWalk(object):
    """GeneWalk object that generates the final output list of significant GO
    terms for each gene in the input list with genes of interest from an
    experiment, for example differentially expressed genes or CRISPR screen
    hits.
    If an input gene is not in the output file, this could have the following
    reasons:
    1) No corresponding HGNC gene symbol, HGNC:ID and/or UniProt:ID could be
    identified. All are required to map genes and assemble their GO
    annotations.  
    2) (if alpha_FDR set to < 1) no GO terms were significant at the
    chosen significance level alpha_FDR.  
    3) (in case of mouse genes) no mapped human ortholog was identified.  
    If a gene is listed in the output file with NaN values in the columns ncon_go 
    and ncon_gene, the gene was not included in the GeneWalk network. 

    Parameters
    ----------
    graph : networkx.MultiGraph
        GeneWalk network for which the statistics are calculated.
    genes : list of dict
        List of gene references for relevant genes.
    nvs : list of dict
        Node vectors for nodes in the graph.
    null_dist : np.array
        Similarity random (null) distribution.
    """
    def __init__(self, graph, genes, nvs, null_dist):
        self.graph = graph
        self.genes = genes
        self.nvs = nvs
        self.srd = null_dist
        self.go_nodes = set(nx.get_node_attributes(self.graph, 'GO'))
        self.gene_nodes = set([g['HGNC_SYMBOL'] for g in self.genes])

    def get_gene_attribs(self, gene):
        """Return an attribute dict for a given gene."""
        if gene['HGNC_SYMBOL'] in self.graph:
            ncon_gene = len(self.graph[gene['HGNC_SYMBOL']])
        else:
            ncon_gene = np.nan
        return {
            'hgnc_symbol': gene['HGNC_SYMBOL'],
            'hgnc_id': gene['HGNC'],
            'ncon_gene': ncon_gene
        }

    def get_go_attribs(self, gene_attribs, nv, alpha_fdr):
        """Return GO entries and their attributes for a given gene."""
        gene_node_id = gene_attribs['hgnc_symbol']
        connected = set(self.graph[gene_node_id]) & self.go_nodes
        go_attribs = []
        pvals = []
        for go_node_id in connected:
            go_attrib = {}
            sim_score = nv.similarity(gene_node_id, go_node_id)
            go_attrib['sim_score'] = sim_score
            go_attrib['go_id'] = go_node_id
            go_attrib['ncon_go'] = len(self.graph[go_node_id])
            go_attrib['go_name'] = self.graph.nodes[go_node_id]['name']
            go_attrib['go_domain'] = \
                self.graph.nodes[go_node_id]['domain'].replace('_', ' ')
            go_attrib['pval'] = self.psim(sim_score)
            pvals.append(go_attrib['pval'])
            go_attribs.append(go_attrib)
        _, qvals = fdrcorrection(pvals, alpha=alpha_fdr, method='indep')
        for idx in range(len(go_attribs)):
            go_attribs[idx]['qval'] = qvals[idx]
        return go_attribs

    def log_stats(self, vals):
        eps = 1e-16
        nreps = len(vals)
        vals = np.asarray(vals)+eps
        g_mean = gmean(vals)-eps
        g_std = gstd(vals)
        return g_mean, g_mean*(g_std**(-1.96/np.sqrt(nreps))), \
            g_mean*(g_std**(1.96/np.sqrt(nreps)))

    def add_empty_row(self, gene, gene_attribs, base_id_type):
        row = [gene_attribs['hgnc_symbol'],
               gene_attribs['hgnc_id'],
               '', '', '',
               gene_attribs['ncon_gene'],
               np.nan, np.nan, np.nan, np.nan,
               np.nan, np.nan, np.nan, np.nan, np.nan]
        if base_id_type == 'mgi_id':
            row = [gene.get('MGI', '')] + row
        elif base_id_type == 'ensembl_id':
            row = [gene.get('ENSEMBL', '')] + row
        elif base_id_type == 'entrez_mouse' or \
             base_id_type == 'entrez_human':
            row = [gene.get('EGID', '')] + row
        return row

    def generate_output(self, alpha_fdr=1, base_id_type='hgnc_symbol'):
        """Main function of GeneWalk object that generates the final
        GeneWalk output table (in csv format).

        Parameters
        ----------
        alpha_fdr : Optional[float]
            Significance level for FDR [0,1] (default=1, i.e. all GO
            terms and their statistics are output). If set to a lower value,
            only connected GO terms with mean padj < alpha_FDR are output.
        base_id_type : Optional[str]
            The type of gene IDs that were the basis of doing the analysis.
            In case of mgi_id or ensembl_id, we prepend a column to the table
            for MGI or ENSEMBL IDs, respectively.
            Default: hgnc_symbol
        """
        rows = []
        for gene in self.genes:
            gene_attribs = self.get_gene_attribs(gene)
            # gene present in GW network
            if not np.isnan(gene_attribs['ncon_gene']):
                all_go_attribs = [self.get_go_attribs(gene_attribs, nv,
                                                      alpha_fdr)
                                  for nv in self.nvs]
                if all_go_attribs:  # gene has GO connections
                    go_attrib_dict = {}
                    for go_attrib_list in all_go_attribs:
                        for go_attribs in go_attrib_list:
                            go_id = go_attribs['go_id']
                            if go_id in go_attrib_dict:
                                go_attrib_dict[go_id].append(go_attribs)
                            else:
                                go_attrib_dict[go_id] = [go_attribs]

                    for go_id, go_attribs in go_attrib_dict.items():
                        mean_padj, low_padj, upp_padj = \
                            self.log_stats([attr['qval']
                                            for attr in go_attribs])
                        mean_pval, low_pval, upp_pval = \
                            self.log_stats([attr['pval']
                                            for attr in go_attribs])
                        mean_sim = np.mean([attr['sim_score']
                                            for attr in go_attribs])
                        sem_sim = (np.std([attr['sim_score']
                                           for attr in go_attribs]) /
                                   np.sqrt(len(self.nvs)))
                        if mean_padj < alpha_fdr or alpha_fdr == 1:
                            row = [gene_attribs['hgnc_symbol'],
                                   gene_attribs['hgnc_id'],
                                   go_attribs[0]['go_name'],
                                   go_attribs[0]['go_id'],
                                   go_attribs[0]['go_domain'],
                                   gene_attribs['ncon_gene'],
                                   go_attribs[0]['ncon_go'],
                                   mean_padj, low_padj, upp_padj,
                                   mean_pval, low_pval, upp_pval,
                                   mean_sim, sem_sim]
                            # If dealing with mouse genes, prepend the MGI ID
                            if base_id_type == 'mgi_id':
                                row = [gene.get('MGI', '')] + row
                            elif base_id_type == 'ensembl_id':
                                row = [gene.get('ENSEMBL', '')] + row
                            elif base_id_type == 'entrez_mouse' or \
                                 base_id_type == 'entrez_human':
                                row = [gene.get('EGID', '')] + row
                            rows.append(row)
                elif alpha_fdr == 1:  # case: no GO connections
                    row = self.add_empty_row(gene, gene_attribs, base_id_type)
                    rows.append(row)
            elif alpha_fdr == 1:  # case: not in graph
                row = self.add_empty_row(gene, gene_attribs, base_id_type)
                rows.append(row)
        header = ['hgnc_symbol', 'hgnc_id',
                  'go_name', 'go_id', 'go_domain',
                  'ncon_gene', 'ncon_go',
                  'mean_padj', 'cilow_padj', 'ciupp_padj',
                  'mean_pval', 'cilow_pval', 'ciupp_pval',
                  'mean_sim',  'sem_sim',
                  ]
        if base_id_type in {'mgi_id', 'ensembl_id', 'entrez_human',
                            'entrez_mouse'}:
            header = [base_id_type] + header

        df = pd.DataFrame.from_records(rows, columns=header)
        df[base_id_type] = df[base_id_type].astype('category')
        df[base_id_type].cat.set_categories(df[base_id_type].unique(),
                                            inplace=True)
        df[['ncon_gene', 'ncon_go']] = \
            df[['ncon_gene', 'ncon_go']].astype('str')
        df = df.sort_values(by=[base_id_type, 'go_domain', 'mean_padj',
                                'mean_sim', 'go_name'],
                            ascending=[True, True, True, False, True])
        return df

    def psim(self, sim):
        """Determine the p-value of the experimental similarity by determining
        its percentile, i.e. the normalized rank, in the null distribution
        with random similarity values.
        """
        rank = np.searchsorted(self.srd, sim)
        pct_rank = float(rank) / len(self.srd)
        pval = 1 - pct_rank
        eps = 1e-16
        if pval < eps:
            pval = eps
        return pval
